#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest unit test Maya settings class
"""

from __future__ import print_function, division, absolute_import

from tpDcc.libs.unittests.core import settings


class MayaUnitTestSettings(settings.BaseUnitTestSettings, object):
    """
    Settings for running unit tests in Maya
    """

    # Specifies whether the standard output and standard error streams are buffered during the test run.
    # Output during a passing test is discarded. Output is echoed normally on test fail or error and is added to
    # the failure messages
    buffer_output = False



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest unit test base class
"""

from __future__ import print_function, division, absolute_import

import os
import shutil
import unittest

from tpDcc import dcc
from tpDcc.libs.python import decorators

from tpDcc.libs.unittests.core import settings


class MetaUnitTestCase(type):
    def __call__(cls, *args, **kwargs):
        as_class = kwargs.get('as_class', True)
        if dcc.is_maya():
            from tpDcc.libs.unittests.dccs.maya import unittest
            if as_class:
                return unittest.MayaUnitTestCase
            else:
                return type.__call__(unittest.MayaUnitTestCase, *args, **kwargs)
        else:
            if as_class:
                return BaseUnitTestCase
            else:
                return type.__call__(BaseUnitTestCase, *args, **kwargs)


class BaseUnitTestCase(unittest.TestCase):
    """
    Base class for unit test cases inside tpDcc.libs.unittest
    New tests are not mandatory to inherit from this TestCase but this derived TestCase contains convenience
    functions to clean up temporary files
    """

    # Keep track of all temporary files that where created so they can be cleaned up after all tests have been run
    files_created = []

    # region Functions
    @classmethod
    def delete_temp_files(cls):
        """
        Delete the temp files in the cache and clear the cache
        """

        # If we don't want to keep temp files around for debugging purposes, delete them when
        # all tests in this tpTestCase have been run
        if settings.UnitTestSettings().delete_files:
            for f in cls.files_created:
                if os.path.exists(f):
                    os.remove(f)
            cls.files_created = []
            if os.path.exists(settings.UnitTestSettings().temp_dir):
                shutil.rmtree(settings.UnitTestSettings().temp_dir)

    @classmethod
    def get_temp_filename(cls, file_name):
        """
        Get a unique file path name in the testing directory
        The file will not be created, that is up to the caller. This file will be deleted when the test are finished
        :param str file_name: A partial path ex: 'directory/someFile.txt'
        :return: str, The full path to the temporary file
        """

        temp_dir = settings.UnitTestSettings().temp_dir
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)
        base_name, ext = os.path.splitext(file_name)
        path = '{0}/{1}{2}'.format(temp_dir, base_name, ext)
        count = 0
        while os.path.exists(path):
            # If the file already exists, add an incremented number
            count += 1
            path = '{0}/{1}{2}{3}'.format(temp_dir, base_name, count, ext)
        cls.files_created.append(path)
        return path

    def assert_list_almost_equal(self, first, second, places=7, msg=None, delta=None):
        """
        Asserts that a list of floating point values are almost equal
        unittest has assertAlmostEqual and assertListEqual but no assertListAlmostEqual
        """

        self.assertEqual(len(first), len(second), msg)
        for a, b in zip(first, second):
            self.assertAlmostEqual(a, b, places, msg, delta)
    # endregion

    # region Override Functions
    def tearDown(self):
        """
        Method that is called before test in an individual class run
        For custom unit tests classes, this method should be override
        """

        super(BaseUnitTestCase, self).tearDown()

    @classmethod
    def tearDownClass(cls):
        """
        A class method called after tests in an individual class have run
        """

        super(BaseUnitTestCase, cls).tearDownClass()


@decorators.add_metaclass(MetaUnitTestCase)
class UnitTestCase(object):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest constant definitions
"""

from __future__ import print_function, division, absolute_import


UNIT_TEST_VAR = 'TPDCC_UNIT_TEST'



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Initialization module for tpDcc-libs-unittests
"""


def init(*args, **kwargs):
    """
    Initializes library
    :param dev: bool, Whether tpDcc-libs-qt is initialized in dev mode or not
    """

    print('Unit Test Library initialized!')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest unit test Maya settings class
"""

from __future__ import print_function, division, absolute_import

import os

from tpDcc.libs.unittests.core import consts, unittestcase, settings

import maya.cmds


class MayaTestCase(unittestcase.BaseUnitTestCase, object):

    """
    Base class for unit test cases run in Maya
    Contains convenience functions to load/unload plug-ins and clean up temporary files
    """

    # Keep track of which plugins were loaded so we can unload them after all tests have been run
    plugins_loaded = set()

    @classmethod
    def load_plugin(cls, plugin):
        """
        Load the given plug-in and saves it to be unloaded when the tpTestCase is finished
        :param str plugin: Plug-in name
        """

        maya.cmds.loadPlugin(plugin, quiet=True)
        cls.plugins_loaded.add(plugin)

    @classmethod
    def unload_plugins(cls):
        """
        Unload any plugins that this test case loaded
        """

        for plugin in cls.plugins_loaded:
            maya.cmds.unloadPlugin(plugin)
        cls.plugins_loaded = []

    def tearDown(self):
        """
        A class method called before tests in an individual class run
        """

        if settings.UnitTestSettings().file_new and consts.UNIT_TEST_VAR not in os.environ.keys():
            # If running tets without the custom runner, like with PyCharm, the file new of the TestResult class
            # is not used so call file new here
            maya.cmds.file(f=True, new=True)

    @classmethod
    def tearDownClass(cls):
        """
        A class method called after tests in an individual class have run
        """

        super(MayaTestCase, cls).tearDownClass()
        cls.delete_temp_files()
        cls.unload_plugins()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest function implementations for Maya
"""

from __future__ import print_function, division, absolute_import

import os
import sys

import maya.cmds
import maya.standalone

from tpDcc.libs.unittests.core import settings, unittestlib


def load_default_unit_tests():
    """
    Loads default unit tests of a specific DCC
    """

    pass


def maya_module_tests():
    """
    Generator function to iterate over all the Maya module tests directories
    """

    maya_modules_paths = list()
    for path in os.environ['MAYA_MODULE_PATH'].split(os.pathsep):
        p = '{0}/tests'.format(path)
        if os.path.exists(p):
            maya_modules_paths.append(p)

    return maya_modules_paths


def run_tests_from_command_line():
    """
    Run the tests in Maya standalone mode
    """

    maya.standalone.initialize()

    # Make sure all paths in PYTHONPATH are also in sys.path
    # When a Maya module is loaded, the scripts folder is added to PYTHONPATH, but it doesn't seem to be added
    # sys.path. So we are unable to import any of the python files that are in the module/scripts folder. To
    # workaround this, we simply add the pats to sys ourselves
    real_sys_path = [os.path.realpath(p) for p in sys.path]
    python_path = os.environ.get('PYTHONPATH', '')
    for p in python_path.split(os.pathsep):
        p = os.path.realpath(p)     # Make sure symbolic links are resolved
        if p not in real_sys_path:
            sys.path.insert(0, p)

    unittestlib.run_tests()

    # Starting Maya 2016, we have to call uninitialize
    if float(maya.cmds.about(v=True)) >= 2016.0:
        maya.standalone.uninitialize()


def set_buffer_output(value):
    """
    Sets whether the standard output and standard error streams are buffered during the test run
    :param bool value: True to buffer output and standard error streams
    """

    settings.UnitTestSettings().bufferOutput = value


def set_file_new(value):
    """
    Sets whether a new file should be created after each test
    :param bool value: True if a new file should be created after each test
    """

    settings.UnitTestSettings().fileNew = value



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest unit test Maya result class
"""

from __future__ import print_function, division, absolute_import

import os
import shutil
import logging

import maya.cmds

from tpDcc.libs.unittests.core import settings, result
from tpDcc.libs.unittests.dccs.maya import scripteditor


class MayaUnitTestResult(result.BaseUnitTestResult, object):
    """
    Customize Maya the test result so we can do things like do a file new between each test and suppress script
    editor output
    """

    def startTestRun(self):
        """
        Called before any tests are run
        """

        super(MayaUnitTestResult, self).startTestRun()

        scripteditor.MayaScriptEditorState.suppress_output()
        if settings.UnitTestSettings().buffer_output:
            # Disable any logging while running tests. By disabling critical, we are disabling logging at all levels
            # below critical as well
            logging.disable(logging.CRITICAL)

    def stopTestRun(self):
        """
        Called after all tests are run
        """

        if settings.UnitTestSettings().buffer_output:
            # Restore logging state
            logging.disable(logging.NOTSET)
        scripteditor.MayaScriptEditorState.restore_output()
        if settings.UnitTestSettings().delete_files and os.path.exists(settings.UnitTestSettings().temp_dir):
            shutil.rmtree(settings.UnitTestSettings().temp_dir)

        super(MayaUnitTestResult, self).stopTestRun()

    def stopTest(self, test):
        """
        Called after an individual test is run
        @param test: TestCase that just ran
        """
        super(MayaUnitTestResult, self).stopTest(test)

        if settings.UnitTestSettings().file_new:
            maya.cmds.file(f=True, new=True)

    def addSuccess(self, test):
        """
        Override the base addSuccess method so we can store a list of the successful tests
        @param test: Testase that sucessfully ran.
        """

        super(MayaUnitTestResult, self).addSuccess(test)
        self._successes.append(test)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest unit test base settings class
"""

from __future__ import print_function, division, absolute_import

import os
import uuid
import tempfile

from tpDcc import dcc
from tpDcc.libs.python import decorators


class MetaUnitTestSettings(type):

    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            if dcc.is_maya():
                from tpDcc.libs.unittests.dccs.maya import settings
                cls._instance = type.__call__(settings.MayaUnitTestSettings, *args, **kwargs)
            else:
                cls._instance = type.__call__(BaseUnitTestSettings, *args, **kwargs)

        return cls._instance


class BaseUnitTestSettings(object):
    """
    Settings for running unit tests
    """

    # Specifies where file generated during tests should be stored
    # Use a uuid subdirectory so test that are running concurrently such as on a build server do not conflict
    # with each other

    # Here we use uuid4() because it generates a random unique UUID, uuid1() generated a random UUID containing the
    # user computer's network address
    temp_dir = os.path.join(tempfile.gettempdir(), '{}_unittest'.format(dcc.get_name()), str(uuid.uuid4()))

    # Controls whether temp files should be deleted after running all tests in the test case
    delete_files = True


@decorators.add_metaclass(MetaUnitTestSettings)
class UnitTestSettings(object):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest unit test base result class
"""

from __future__ import print_function, division, absolute_import

import os
import unittest

from tpDcc import dcc
from tpDcc.libs.python import decorators
from tpDcc.libs.unittests.core import consts


class MetaUnitTestResult(type):

    def __call__(cls, *args, **kwargs):
        as_class = kwargs.pop('as_class', False)
        if dcc.is_maya():
            from tpDcc.libs.unittests.dccs.maya import result
            if as_class:
                return result.MayaUnitTestResult
            else:
                return type.__call__(result.MayaUnitTestResult, *args, **kwargs)
        else:
            if as_class:
                return BaseUnitTestResult
            else:
                return type.__call__(BaseUnitTestResult, *args, **kwargs)


class BaseUnitTestResult(unittest.TextTestResult):
    """
    Base test result class
    """

    def __init__(self, stream, descriptions, verbosity):
        super(BaseUnitTestResult, self).__init__(stream, descriptions, verbosity)
        self._successes = list()

    @property
    def successes(self):
        return self._successes

    def startTestRun(self):
        """
        Called before any tests are run
        """

        super(BaseUnitTestResult, self).startTestRun()

        # Create an environment variable that specifies tests are being run through the custom runner
        os.environ[consts.UNIT_TEST_VAR] = '1'

    def stopTestRun(self):
        """
        Called after all tests are run
        """

        del os.environ[consts.UNIT_TEST_VAR]

        super(BaseUnitTestResult, self).stopTestRun()

    def stopTest(self, test):
        """
        Called after an individual test is run
        @param test: TestCase that just ran
        """
        super(BaseUnitTestResult, self).stopTest(test)

    def addSuccess(self, test):
        """
        Override the base addSuccess method so we can store a list of the successful tests
        @param test: Testase that sucessfully ran.
        """

        super(BaseUnitTestResult, self).addSuccess(test)
        self._successes.append(test)


@decorators.add_metaclass(MetaUnitTestResult)
class UnitTestResult(object):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains Maya script editor state class implementation
"""

from __future__ import print_function, division, absolute_import

import maya.cmds

from tpDcc.libs.unittests.core import settings


class MayaScriptEditorState (object):
    """
    Provides methods to suppress and restore script editor output
    """

    # Used to restore logging states in the script editor
    suppress_results = None
    suppress_errors = None
    suppress_warnings = None
    suppress_info = None

    @classmethod
    def suppress_output(cls):
        """
        Hides all script editor output
        """

        if settings.UnitTestSettings().buffer_output:
            cls.suppress_results = maya.cmds.scriptEditorInfo(query=True, suppressResults=True)
            cls.suppress_errors = maya.cmds.scriptEditorInfo(query=True, suppressErrors=True)
            cls.suppress_warnings = maya.cmds.scriptEditorInfo(query=True, suppressWarnings=True)
            cls.suppress_info = maya.cmds.scriptEditorInfo(query=True, supressInfo=True)
            maya.cmds.scriptEditorInfo(
                edit=True, suppressResults=True, suppressInfo=True, suppressWarnings=True, suppressErrors=True)

    @classmethod
    def restore_output(cls):
        """
        Restores the script editor output settings to their original values
        """

        if None not in {cls.suppress_results, cls.suppress_errors, cls.suppress_warnings, cls.suppress_info}:
            maya.cmds.scriptEditorInfo(
                edit=True, suppressResults=cls.suppress_results, suppressInfo=cls.suppress_info,
                suppressWarnings=cls.suppress_warnings, suppressErrors=cls.suppress_errors)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-libs-unittest function core implementations
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import logging
import unittest

from tpDcc import dcc
from tpDcc.core import library, reroute
from tpDcc.libs.python import python

from tpDcc.libs.unittests.core import settings, result

LIB_ID = 'tpDcc-libs-unittests'
LIB_ENV = LIB_ID.replace('-', '_').upper()

LOGGER = logging.getLogger('tpDcc-libs-unittests')


class UnitTestsLib(library.DccLibrary, object):
    def __init__(self, *args, **kwargs):
        super(UnitTestsLib, self).__init__(*args, **kwargs)

    @classmethod
    def config_dict(cls, file_name=None):
        base_tool_config = library.DccLibrary.config_dict(file_name=file_name)
        tool_config = {
            'name': 'Unit Tests Library',
            'id': LIB_ID,
            'supported_dccs': {'maya': ['2017', '2018', '2019', '2020']},
            'tooltip': 'Library to manage unit tests in a DCC agnostic way'
        }
        base_tool_config.update(tool_config)

        return base_tool_config


@reroute.reroute_factory(LIB_ID, 'unittestlib')
def load_default_unit_tests():
    """
    Loads default unit tests of a specific DCC
    """

    return list()


@reroute.reroute_factory(LIB_ID, 'unittestlib')
def run_tests_from_command_line():
    """
    Run the tests in Maya standalone mode
    """

    raise NotImplementedError('run test from command line not implemented for current DCC: {}'.format(dcc.get_name()))


def run_tests(directories=None, test=None, test_suite=None):

    """
    Run all the tests in the given paths
    :param list(str) directories: A generator or list of paths containing tests to run
    :param str test: Optional name of a specific test to run
    :param unittest.TestSuite test_suite: Optional TestSuite to run. If omitted, a TestSuite will be generated
    :return: None
    """

    if test_suite is None:
        test_suite = get_tests(directories, test)

    runner = unittest.TextTestRunner(verbosity=2, resultclass=result.BaseUnitTestResult)
    runner.failfast = False
    try:
        runner.buffer = settings.UnitTestSettings().buffer_output
    except Exception:
        pass
    runner.run(test_suite)


def get_tests(directories=None, test=None, test_suite=None):
    """
    Get a UnitTests containing all the desired tests
    :param list(str) directories:  Optional list of directories with which to search for tests. If omitted, use all
    "tests" directories of the modules found in specific app modules path (e.g, maya - MAYA_MODULE_PATH)
    :param str test: Optional test path to find a specific test such as 'test_myTest.SomeTestCase.test_function'
    :param unittest.TestSuite test_suite: Optional unittest.TestSuite to add the discovered test to. If omitted a new
    TestSuite will be created
    :return: unittest.TestSuite
    """

    directories = python.force_list(directories)
    directories.extend(python.force_list(load_default_unit_tests()))

    # Populate a TestSuite with all the tests
    if test_suite is None:
        test_suite = unittest.TestSuite()

    if test:
        # Find the specified test to run
        directories_added_to_path = [p for p in directories if add_to_path(p)]
        discovered_suite = unittest.TestLoader().loadTestsFromName(test)
        if discovered_suite.countTestCases():
            test_suite.addTests(discovered_suite)
    else:
        # Find all tests to run
        directories_added_to_path = []
        for p in directories:
            discovered_suite = unittest.TestLoader().discover(p)
            if discovered_suite.countTestCases():
                new_test_suite = unittest.TestSuite()
                for suites in discovered_suite:
                    for suite in suites._tests:
                        new_test_suite.addTest(suite)
                test_suite.addTest(new_test_suite)

    # Remove the added paths
    for path in directories_added_to_path:
        sys.path.remove(path)

    return test_suite


def set_temp_dir(directory):

    """
    Sets where files generated from tests should be stored
    :param str directory: A directory path
    :return: None
    """

    if os.path.exists(directory):
        settings.UnitTestSettings().temp_dir = directory
    else:
        raise RuntimeError('{0} does not exists.'.format(directory))


def set_delete_files(value):

    """
    Sets whether temp files should be delted after running all tests in a test case
    :param bool value: True to delete files registered with a TestCase
    :return: None
    """

    settings.UnitTestSettings().delete_files = value


def add_to_path(path):

    """
    Add the specified path to the system path
    :param str path: Path to add
    :return: bool True if the path was added. False if path does not exist or path was already in sys.path
    """

    if os.path.exists(path) and path not in sys.path:
        sys.path.insert(0, path)
        return True

    return False
